@@ -1,12 +1,21 @@
 package Babble::Match;
 
-use PPR::X;
+use Babble::Grammar;
+use Babble::SymbolGenerator;
 use Mu;
 use re 'eval';
 
 ro 'top_rule';
 rwp 'text';
-lazy 'grammar_regexp' => sub { $PPR::X::GRAMMAR };
+
+lazy 'grammar' => sub { Babble::Grammar->new }
+  => handles => [ 'grammar_regexp' ];
+
+lazy 'symbol_generator' => sub {
+  $_[0]->can('parent')
+    ? $_[0]->parent->symbol_generator
+    : Babble::SymbolGenerator->new
+  } => handles => [ 'gensym' ];
 
 lazy top_re => sub {
   my ($self) = @_;
@@ -48,6 +57,15 @@ lazy submatches => sub {
   return \%submatches;
 };
 
+sub subtexts {
+  my ($self, @names) = @_;
+  unless (@names) {
+    my %s = %{$self->submatches};
+    return +{ map +( $_ => $s{$_}->text ), keys %s };
+  }
+  map +($_ ? $_->text : undef), @{$self->submatches}{@names};
+}
+
 sub _rule_to_re {
   my $re = $_[1];
   return "(?&Perl${re})" unless ref($re);
@@ -62,9 +80,9 @@ sub is_valid {
 sub match_positions_of {
   my ($self, $of) = @_;
   our @F;
-  my $wrapped = qr{(?(DEFINE)
-    (?<Perl${of}>((?&PerlStd${of}))(?{ push @F, [ pos() - length($^N), length($^N) ] }))
-  ) ${\$self->grammar_regexp}}x;
+  my $wrapped = $self->grammar->clone->extend_rule(
+                  $of => sub { '('.$_[0].')'.'(?{ push @Babble::Match::F, [ pos() - length($^N), length($^N) ] })' }
+                )->grammar_regexp;
   my @found = do {
     local @F;
     local $_ = $self->text;
@@ -82,7 +100,7 @@ sub each_match_of {
   while (my $f = shift @found) {
     my $match = substr($self->text, $f->[0], $f->[1]);
     my $obj = Babble::SubMatch->new(
-                top_rule => $self->_rule_to_re($of),
+                top_rule => $of,
                 start => $f->[0],
                 text => $match,
                 parent => $self,
@@ -104,13 +122,11 @@ sub each_match_of {
 sub each_match_within {
   my ($self, $within, $rule, $call) = @_;
   my $match_re = $self->_rule_to_re($rule);
-  my $extend_grammar = qq{
-    (?(DEFINE)
-      (?<PerlBabbleInnerMatch>(?<PerlStdBabbleInnerMatch> ${match_re}))
-      (?<Perl${within}> (?&PerlBabbleInnerMatch) | (?&PerlStd${within}))
-    )
-  };
-  local $self->{grammar_regexp} = join "\n", $extend_grammar, $self->grammar_regexp;
+  my $extend_grammar = $self->grammar->clone;
+  $extend_grammar->add_rule(
+    BabbleInnerMatch => $match_re,
+  )->augment_rule($within => '(?&PerlBabbleInnerMatch)');
+  local $self->{grammar} = $extend_grammar;
   $self->each_match_of(BabbleInnerMatch => sub {
     $_[0]->{top_rule} = $rule; # intentionally hacky, should go away (or rwp) later
     $call->($_[0]);
@@ -124,6 +140,7 @@ sub replace_substring {
   substr($text, $start, $length, $replace);
   $self->_set_text($text);
   foreach my $submatch (values %{$self->submatches}) {
+    next unless defined $submatch;
     if ($submatch->start > $start) {
       $submatch->{start} += length($replace) - $length;
     }
@@ -131,4 +148,33 @@ sub replace_substring {
   return $self;
 }
 
+sub remove_use_argument {
+  my ($self, $use, $argument, $keep_empty) = @_;
+  $self->each_match_within(
+    UseStatement =>
+    [ "use\\s+${use}\\s+", [ explist => '.*?' ], ';' ],
+    sub {
+      my ($m) = @_;
+      my $explist = $m->submatches->{explist};
+      return unless my @explist_names = eval $explist->text;
+      my @remain = grep $_ ne $argument, @explist_names;
+      return unless @remain < @explist_names;
+      unless (@remain) {
+        ($keep_empty ? $explist : $m)->replace_text('');
+        return;
+      }
+      $explist->replace_text('qw('.join(' ', @remain).')');
+    }
+  );
+}
+
+sub remove_use_statement {
+  my ($self, $use) = @_;
+  $self->each_match_within(
+    UseStatement =>
+    [ "use\\s+${use}.*?;" ],
+    sub { shift->replace_text('') },
+  );
+}
+
 1;
